#pragma once
#ifndef __LIBLSS_PHYSICS_FORWARDS_PATCH_MODEL_HPP
#  define __LIBLSS_PHYSICS_FORWARDS_PATCH_MODEL_HPP

#  include "libLSS/physics/sum.hpp"
#  include "libLSS/physics/forward_model.hpp"

#  include "libLSS/physics/forwards/registry.hpp"

LIBLSS_REGISTER_FORWARD_DECL(PATCH_MODEL);

#endif
