/*+
    ARES/HADES/BORG Package -- ./extra/borg/libLSS/borg_version.hpp
    Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
    
+*/
#ifndef __LIBLSS_BORG_VERSION_HPP
#define __LIBLSS_BORG_VERSION_HPP

#include <string>

namespace LibLSS {

  extern const std::string BORG_GIT_VERSION;

}

#endif
