/*+
    ARES/HADES/BORG Package -- ./extra/python/python/pyvelocity.hpp
    Copyright (C) 2019 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
    
+*/
#ifndef __LIBLSS_PYTHON_FORWARD_VELOCITY_HPP
#  define __LIBLSS_PYTHON_FORWARD_VELOCITY_HPP
#  pragma once

#  include <memory>
#  include <exception>
#  include <boost/format.hpp>
#  include "libLSS/physics/forward_model.hpp"
#  include "libLSS/physics/velocity/velocity.hpp"
#  include "libLSS/physics/velocity/velocity_cic.hpp"
#  include "libLSS/physics/velocity/velocity_sic.hpp"

namespace LibLSS {} // namespace LibLSS

#endif
// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2019
