/*+
    ARES/HADES/BORG Package -- ./extra/hades/libLSS/physics/likelihoods/base.cpp
    Copyright (C) 2018 Guilhem Lavaux <guilhem.lavaux@iap.fr>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
    
+*/
#include "libLSS/physics/likelihoods/base.hpp"

namespace LibLSS {
  namespace Likelihood {
    const std::string MPI = "MPI", COLOR_MAP = "COLOR_MAP", GRID = "GRID",
                      MPI_GRID = "MPI_GRID", GRID_LENGTH = "GRID_LENGTH", DATA_GRID = "DATA_GRID";
  }
} // namespace LibLSS

// ARES TAG: authors_num = 1
// ARES TAG: name(0) = Guilhem Lavaux
// ARES TAG: email(0) = guilhem.lavaux@iap.fr
// ARES TAG: year(0) = 2018
