/*+
    ARES/HADES/BORG Package -- ./libLSS/mcmc/state_element.cpp
    Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
    
+*/
#include <H5Cpp.h>
#include <CosmoTool/hdf5_array.hpp>
#include "state_element.hpp"

using namespace LibLSS;

StateElement::~StateElement()
{
}

