/*+
    ARES/HADES/BORG Package -- ./extra/hades/libLSS/samplers/core/simpleLikelihood.cpp
    Copyright (C) 2014-2020 Guilhem Lavaux <guilhem.lavaux@iap.fr>
    Copyright (C) 2009-2020 Jens Jasche <jens.jasche@fysik.su.se>

    Additional contributions from:
       Guilhem Lavaux <guilhem.lavaux@iap.fr> (2023)
    
+*/
#include "libLSS/samplers/core/simpleLikelihood.hpp"

using namespace LibLSS;

SimpleLikelihood::SimpleLikelihood(LikelihoodInfo const &info) {}

SimpleLikelihood::~SimpleLikelihood() {}